/******************************************************************************/
// dwfbfile.c - library for r/w of DK2 DWFB/WAD files.
/******************************************************************************/
// Author:   Tomasz Lis
// Created:  15 Jul 2008

// Purpose:
//   Allows reading/writing DK2 DWFB files (WADs).

// Comment:
//   Thanks to Trass3r for his decompression algorithm.

//Copying and copyrights:
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
/******************************************************************************/

#include "dwfbfile.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include "lbfileio.h"

const char *dwfb_magic="DWFB";

/*
 * Displays simple error message.
 */
short dwfb_error(const char *msg)
{
    printf("DWFB Support: %s\n",msg);
    return 1;
}

/*
 * Displays formatted error message.
 */
short dwfb_ferror(const char *format, ...)
{
    static char errmessage[255];
    va_list val;
    va_start(val, format);
    vsprintf(errmessage,format,val);
    va_end(val);
    short result;
    result=dwfb_error(errmessage);
    return result;
}

/*
 * Prints all information stored in DWFB_Header.
 */
short dwfb_print_header(const struct DWFB_Header *hdr)
{
    printf("Magic: \"%c%c%c%c\", version %d\n",hdr->magic[0],hdr->magic[1],
        hdr->magic[2],hdr->magic[3],hdr->version);
    printf("Entries %3d, unknown %d\n",hdr->dirnum,hdr->unknown);
    printf("Names block offset %07x, length %2d\n",hdr->name_offs,hdr->name_len);
    printf("Padding: ");
    int i;
    for (i=0;i<16;i++)
      printf("%02x ",(int)(hdr->padding[i]));
    printf("\n         ");
    for (i=16;i<32;i++)
      printf("%02x ",(int)(hdr->padding[i]));
    printf("\n         ");
    for (i=32;i<48;i++)
      printf("%02x ",(int)(hdr->padding[i]));
    printf("\n         ");
    for (i=48;i<64;i++)
      printf("%02x ",(int)(hdr->padding[i]));
    printf("\n");
    return 1;
}

/*
 * Prints all information stored in DWFB_Entry.
 */
short dwfb_print_entry(const struct DWFB_Entry *entry)
{
    printf("Name offset %07x, length %2d, content \"%s\"\n",entry->name_offs,entry->name_len,entry->name);
    printf("Body offset %07x, length %6d, orig. length %d\n",
        entry->body_offs,entry->body_len,entry->original_len);
    printf("Compression %d, unknown %d\n",entry->store_method,entry->unknown);
    printf("Padding: ");
    int i;
    for (i=0;i<12;i++)
      printf("%02x ",(int)(entry->padding[i]));
    printf("\n");
    return 1;
}

/*
 * Loads single DWFB name data block from disk file.
 * Note that this is low-level function; use dwfb_fread_entry_name instead.
 * Returns amount of bytes readed, or negative value on error.
 */
long dwfb_fread_strdata(char **data,FILE *fp,long offset,long length,short flags)
{
  if ((fp==NULL)||(offset<SIZEOF_DWFB_Header)||(length<1))
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("Bad initial variables for name reading");
      (*data)=strdup("");
      return -1;
  }
  if (fseek(fp,offset,SEEK_SET) != 0)
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("%s while seeking for name",strerror(errno));
      (*data)=strdup("");
      return -1;
  }
  (*data)=malloc(length+1);
  if ((*data)==NULL)
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("Can't allocate memory to store name");
      (*data)=strdup("");
      return -1;
  }
  long nread=fread((*data),1,length,fp);
  if (nread != length)
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("%s while reading name",strerror(errno));
  }
  (*data)[length]='\0';
  return nread;
}

/*
 * Loads single, RAW DWFB data block from disk file.
 * No processing is performed to the returned block - RAW read only.
 * Note that this is low-level function; use dwfb_fread_entry_body instead.
 * Returns amount of bytes readed, or negative value on error.
 */
long dwfb_fread_rawdata(unsigned char **data,FILE *fp,long offset,long length,short flags)
{
  if ((fp==NULL)||(offset<SIZEOF_DWFB_Header)||(length<0))
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("Bad initial variables for data reading");
      (*data)=NULL;
      return -1;
  }
  if (length==0)
  {
      (*data)=NULL;
      return 0;
  }
  if (fseek(fp,offset,SEEK_SET) != 0)
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("%s while seeking for data",strerror(errno));
      (*data)=NULL;
      return -1;
  }
  (*data)=malloc(length);
  if ((*data)==NULL)
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("Can't allocate memory to store data");
      return -1;
  }
  long nread=fread((*data),1,length,fp);
  if (nread != length)
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("%s while reading data",strerror(errno));
  }
  return nread;
}

/*
 * Properly loads single DWFB data block from disk file.
 * The block is decompressed and stored in body pointer of the DWFB_Entry.
 * Returns amount of bytes readed, or negative value on error.
 */
long dwfb_fread_entry_body(struct DWFB_File *dwfbfile,struct DWFB_Entry *entry,short flags)
{
  unsigned char *data;
  long nbytes;
  if ( (entry==NULL) || (dwfbfile==NULL) )
    return -1;
  if (entry->body!=NULL)
      free(entry->body);
  nbytes=dwfb_fread_rawdata(&data,dwfbfile->fp,entry->body_offs,entry->body_len,flags);
  switch (entry->store_method)
  {
  case DWFB_COMPRESS_NONE:
      entry->body=data;
      break;
  case DWFB_COMPRESS_MTZ4:
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_error("Compression method 4 not supported");
      entry->body=data;
      return -1;
  default:
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("File uses unknown compression method %d",entry->store_method);
      entry->body=data;
      return -1;
  }
  //TODO: handle compression
  if (nbytes>0)
    dwfbfile->iosize+=nbytes;
  return nbytes;
}

/*
 * Properly loads single DWFB entry name from disk file.
 * Returns amount of bytes readed, or negative value on error.
 */
long dwfb_fread_entry_name(struct DWFB_File *dwfbfile,struct DWFB_Entry *entry,short flags)
{
  char *data;
  long nbytes;
  if ( (entry==NULL) || (dwfbfile==NULL) )
    return -1;
  if (entry->name!=NULL)
      free(entry->name);
  nbytes=dwfb_fread_strdata(&data,dwfbfile->fp,entry->name_offs,entry->name_len,flags);
  entry->name=data;
  if (nbytes>0)
    dwfbfile->iosize+=nbytes;
  return nbytes;
}

/*
 * Loads DWFB header from current position of disk file.
 * Requires the DWFB_Header to be allocated before.
 * Returns SIZEOF_DWFB_Header on success.
 */
long dwfb_fread_header_simp(struct DWFB_Header *header,FILE *fp)
{
  long nread=0;
  nread+=fread(&(header->magic),1,4,fp);
  header->version=read_int32_le_file(fp);
  nread+=4;
  nread+=fread(&(header->padding),1,64,fp);
  header->dirnum=read_int32_le_file(fp);
  nread+=4;
  header->name_offs=read_int32_le_file(fp);
  nread+=4;
  header->name_len=read_int32_le_file(fp);
  nread+=4;
  header->unknown=read_int32_le_file(fp);
  nread+=4;
  return nread;
}

/*
 * Loads DWFB entry from current position of disk file.
 * Requires the DWFB_Entry to be allocated before.
 * Returns SIZEOF_DWFB_Entry on success.
 */
long dwfb_fread_entry_simp(struct DWFB_Entry *entry,FILE *fp)
{
  long nread=0;
  entry->unknown=read_int32_le_file(fp);
  nread+=4;
  entry->name_offs=read_int32_le_file(fp);
  nread+=4;
  entry->name_len=read_int32_le_file(fp);
  nread+=4;
  entry->body_offs=read_int32_le_file(fp);
  nread+=4;
  entry->body_len=read_int32_le_file(fp);
  nread+=4;
  entry->store_method=read_int32_le_file(fp);
  nread+=4;
  entry->original_len=read_int32_le_file(fp);
  nread+=4;
  nread+=fread(&(entry->padding),1,12,fp);
  return nread;
}

/*
 * Loads single DWFB entry from disk file.
 */
short dwfb_fread_entry(struct DWFB_Entry **entry,unsigned long idx,FILE *fp,long offset,short flags)
{
  if ((entry==NULL)||(fp==NULL)||(offset<SIZEOF_DWFB_Header))
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("Bad initial data for entry %lu reading",idx);
      return 0;
  }
  if (fseek(fp,offset,SEEK_SET) != 0)
  {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_ferror("%s while seeking for entry %lu",strerror(errno),idx);
      return 0;
  }
  (*entry)=malloc(sizeof(struct DWFB_Entry));
  if ((*entry)==NULL)
  {
    if (flags&DWFBFLAG_VERBOSE)
      dwfb_ferror("Cannot allocate DWFB_Entry %lu memory",idx);
    return 0;
  }
  long nread=dwfb_fread_entry_simp(*entry,fp);
  if (nread != SIZEOF_DWFB_Entry)
  {
    if (flags&DWFBFLAG_VERBOSE)
      dwfb_ferror("%s while reading entry %lu",strerror(errno),idx);
    free(*entry);
    (*entry)=NULL;
    return 0;
  }
  (*entry)->name=NULL;
  (*entry)->body=NULL;
  return 1;
}

/*
 * Returns name of the given DWFB_Entry. If the name is not loaded yet,
 * loads it from disk file.
 */
char *dwfb_entry_name(struct DWFB_File *dwfbfile,struct DWFB_Entry *entry,short flags)
{
    if (entry==NULL)
    {
        if (flags&DWFBFLAG_VERBOSE)
          dwfb_ferror("Can't get name of null entry");
        return "";
    }
    if (entry->name==NULL)
        dwfb_fread_entry_name(dwfbfile,entry,flags);
    return entry->name;
}

/*
 * Returns name of the given DWFB_Entry. If the name is not loaded yet,
 * loads it from disk file.
 */
char *dwfb_entryidx_name(struct DWFB_File *dwfbfile,unsigned long idx,short flags)
{
     if ((dwfbfile==NULL)||(dwfbfile->entries==NULL)) return "";
     if (idx >= dwfbfile->header.dirnum)
     {
        if (flags&DWFBFLAG_VERBOSE)
          dwfb_ferror("Can't get name of outranged entry");
        return "";
     }
     return dwfb_entry_name(dwfbfile,dwfbfile->entries[idx],flags);
}

/*
 * Returns data block of the given DWFB_Entry. If the data is not loaded yet,
 * loads it from disk file.
 */
char *dwfb_entry_body(struct DWFB_File *dwfbfile,struct DWFB_Entry *entry,short flags)
{
    if (entry==NULL)
    {
        if (flags&DWFBFLAG_VERBOSE)
          dwfb_ferror("Can't get name of null entry");
        return "";
    }
    if (entry->body==NULL)
        dwfb_fread_entry_body(dwfbfile,entry,flags);
    return entry->body;
}


/*
 * Returns name of the given DWFB_Entry. If the name is not loaded yet,
 * loads it from disk file.
 */
char *dwfb_entryidx_body(struct DWFB_File *dwfbfile,unsigned long idx,short flags)
{
     if ((dwfbfile==NULL)||(dwfbfile->entries==NULL)) return NULL;
     if (idx >= dwfbfile->header.dirnum)
     {
        if (flags&DWFBFLAG_VERBOSE)
          dwfb_ferror("Can't get body of outranged entry");
        return NULL;
     }
     return dwfb_entry_body(dwfbfile,dwfbfile->entries[idx],flags);
}

/*
 * Loads DWFB file from given filename and creates structure for maintaining it.
 */
struct DWFB_File *dwfb_open(char *fname,short flags)
{
  struct DWFB_File *dwfbfile;
  FILE *fp;
  dwfbfile=malloc(sizeof(struct DWFB_File));
  if (dwfbfile==NULL)
  {
    if (flags&DWFBFLAG_VERBOSE)
      dwfb_error("Cannot allocate DWFB_File memory");
    return NULL;
  }
  fp=fopen(fname,"rb");
  dwfbfile->fp=fp;
  if (fp==NULL)
  {
    if (flags&DWFBFLAG_VERBOSE)
      dwfb_error(strerror(errno));
    free(dwfbfile);
    return NULL;
  }
  int nread;
  dwfbfile->iosize=0;

  nread=dwfb_fread_header_simp(&(dwfbfile->header),fp);
  dwfbfile->iosize+=nread;
  if (nread != SIZEOF_DWFB_Header)
  {
    if (flags&DWFBFLAG_VERBOSE)
      dwfb_ferror("%s when reading header",strerror(errno));
    free(dwfbfile);
    fclose(fp);
    return NULL;
  }

  if (!(flags&DWFBFLAG_IGNORE_BADHDR))
    if (strncmp(dwfbfile->header.magic,dwfb_magic,4) != 0)
    {
      if (flags&DWFBFLAG_VERBOSE)
        dwfb_error("File has no DWFB header");
      free(dwfbfile);
      fclose(fp);
      return NULL;
    }
  dwfbfile->disksize=file_length_opened(fp);
  unsigned long entries_size=dwfbfile->header.dirnum*sizeof(struct DWFB_Entry *);
  dwfbfile->entries=(struct DWFB_Entry **)malloc(entries_size);
  if (dwfbfile->entries == NULL)
  {
    if (flags&DWFBFLAG_VERBOSE)
      dwfb_error("Cannot allocate DWFB_Entries");
    free(dwfbfile);
    fclose(fp);
    return NULL;
  }
  memset(dwfbfile->entries,0,entries_size);
  int i;
  for (i=0;i<dwfbfile->header.dirnum;i++)
  {
      short result;
      result=dwfb_fread_entry(&(dwfbfile->entries[i]),i,fp,SIZEOF_DWFB_Header+i*SIZEOF_DWFB_Entry,flags);
      if (result!=1)
      {
        dwfb_close(dwfbfile,0);
        fclose(fp);
        return NULL;
      }
      dwfbfile->iosize+=SIZEOF_DWFB_Entry;
  }
  if (flags&DWFBFLAG_FORCE_LOADNAME)
    for (i=0;i<dwfbfile->header.dirnum;i++)
    {
        struct DWFB_Entry *entry=dwfbfile->entries[i];
        long nbytes=dwfb_fread_entry_name(dwfbfile,entry,flags);
    }
  if (flags&DWFBFLAG_FORCE_LOADDATA)
    for (i=0;i<dwfbfile->header.dirnum;i++)
    {
        struct DWFB_Entry *entry=dwfbfile->entries[i];
        long nbytes=dwfb_fread_entry_body(dwfbfile,entry,flags);
    }
  if ( (flags&DWFBFLAG_FORCE_LOADDATA) && (flags&DWFBFLAG_FORCE_LOADNAME) )
  {
    fclose(fp);
    dwfbfile->fp=NULL;
  }
  return dwfbfile;
}

short dwfb_close(struct DWFB_File *dwfbfile,short flags)
{
  if (dwfbfile==NULL) return 0;
  if (dwfbfile->entries!=NULL)
  {
    int i;
    for (i=0;i<dwfbfile->header.dirnum;i++)
    {
      if (dwfbfile->entries[i]!=NULL)
      {
        free(dwfbfile->entries[i]->body);
        free(dwfbfile->entries[i]->name);
        free(dwfbfile->entries[i]);
      }
    }
    free(dwfbfile->entries);
  }
  if (dwfbfile->fp != NULL)
    fclose(dwfbfile->fp);
  free(dwfbfile);
  return 1;
}

long dwfb_decomp4_alg(unsigned char *dest, unsigned long dest_len, const unsigned char *src, unsigned long src_len)
{
    if ( (src == NULL) || (dest == NULL) || (src_len<4) )
        return -1;

    unsigned long i=0,j=0;
    if (src[i++] & 1)
        i+=3;
    i++; // skip second byte
    // decompressed size packed into 3 bytes
    unsigned long decsize = (src[i]<<16) + (src[i+1]<<8) + src[i+2];
    i+=3;
    unsigned char flag; // the flag byte read at the beginning of each main loop iteration
    unsigned long counter; // counter for all loops
    short finished;
    finished=0;
    while (!finished)
    {
        if (i>=src_len)
            break;
        flag = src[i++]; // get flag byte
        if (!(flag & 0x80))
        {
            unsigned char tmp = src[i++];
            counter = flag&3; // mod 4
            while (counter--) // copy literally
            {
                dest[j]=src[i++];
                j++;
            }
            unsigned long k=j; // get the destbuf position
            k -= ((unsigned long)flag&0x60)<<3;
            k -= tmp;
            k--;

            counter=((((unsigned long)flag)>>2)&7)+2;
            do
            {
                dest[j]=dest[k++];
                j++;
            } while(counter--); // correct decrement
        } else
        if (!(flag & 0x40))
        {
            unsigned char tmp = src[i++];
            unsigned char tmp2 = src[i++];
            counter=((unsigned long)tmp)>>6;
            while(counter--) // copy literally
            {
                dest[j]=src[i++];
                j++;
            }
            unsigned long k=j;
            k -= ((unsigned long)tmp&0x3F)<<8;
            k -= tmp2;
            k--;
            counter=(flag&0x3F)+3;
            do
            {
                dest[j]=dest[k++];
                j++;
            } while(counter--); // correct postfix decrement
        } else
        if (!(flag & 0x20))
        {
            unsigned char localtemp=src[i++];
            unsigned char tmp2=src[i++];
            unsigned char tmp3=src[i++];
            counter=flag&3;
            while(counter--) // copy literally
            {
                dest[j]=src[i++];
                j++;
            }
            unsigned long k=j;
            k -= ((unsigned long)flag&0x10)<<12;
            k -= localtemp<<8;
            k -= tmp2;
            k--;
            counter=tmp3+(((unsigned long)flag&0x0C)<<6)+4;
            do
            {
                dest[j]=dest[k++];
                j++;
            } while (counter--); // correct
        } else
        {
            counter = (flag & 0x1F)*4 + 4;
            if (counter > 0x70)
            {
                finished=1;
                // prepare to copy the last bytes
                counter = flag&3;
            }
            while (counter--) // copy literally
            {
                dest[j]=src[i++];
                j++;
            }
        }
    } // of while()
    if (!finished)
      return -1;
    return decsize;
}

